"use strict";

Object.defineProperty(exports, "__esModule", {
    value: true
});

var _regenerator = require("babel-runtime/regenerator");

var _regenerator2 = _interopRequireDefault(_regenerator);

var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");

var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);

var _createClass2 = require("babel-runtime/helpers/createClass");

var _createClass3 = _interopRequireDefault(_createClass2);

var _promise = require("babel-runtime/core-js/promise");

var _promise2 = _interopRequireDefault(_promise);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var __awaiter = undefined && undefined.__awaiter || function (thisArg, _arguments, P, generator) {
    return new (P || (P = _promise2.default))(function (resolve, reject) {
        function fulfilled(value) {
            try {
                step(generator.next(value));
            } catch (e) {
                reject(e);
            }
        }
        function rejected(value) {
            try {
                step(generator["throw"](value));
            } catch (e) {
                reject(e);
            }
        }
        function step(result) {
            result.done ? resolve(result.value) : new P(function (resolve) {
                resolve(result.value);
            }).then(fulfilled, rejected);
        }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
};
/** Class representing a semaphore
 * Semaphores are initialized with a number of permits that get aquired and released
 * over the lifecycle of the Semaphore. Functions can wait and stop executing until
 * a permit gets released (the method for which is called signal for historical reasons.)
 *
 * Locks that only allow one execution of a critical section are a special case of
 * Semaphores. To construct a lock, initialize a Semaphore with a permit count of 1.
 *
 * This Semaphore class is implemented with the help of promises that get returned
 * by functions that wait for permits to become available. This makes it possible
 * to use ES7 async/awaits to synchronize your code.
*/
/** Class representing a semaphore
 * Semaphores are initialized with a number of permits that get aquired and released
 * over the lifecycle of the Semaphore. Functions can wait and stop executing until
 * a permit gets released (the method for which is called signal for historical reasons.)
 *
 * Locks that only allow one execution of a critical section are a special case of
 * Semaphores. To construct a lock, initialize a Semaphore with a permit count of 1.
 *
 * This Semaphore class is implemented with the help of promises that get returned
 * by functions that wait for permits to become available. This makes it possible
 * to use ES7 async/awaits to synchronize your code.
*/
var Semaphore = function () {
    /**
     * Creates a semaphore.
     * @param {number} permits The number of permits, i.e. things being allowed to run in parallel.
     * To create a lock that only lets one thing run at a time, set this to 1.
     * This number can also be negative.
     */
    function Semaphore(permits) {
        (0, _classCallCheck3.default)(this, Semaphore);

        this.permits = permits;
        this.promiseResolverQueue = [];
    }
    /**
     * Returns a promise used to wait for a permit to become available. This method is supposed
     * to be awaited on.
     * @return {Promise} A promise that gets resolved when execution is allowed to proceed.
     */


    (0, _createClass3.default)(Semaphore, [{
        key: "wait",
        value: function wait() {
            return __awaiter(this, void 0, void 0, _regenerator2.default.mark(function _callee() {
                var _this = this;

                return _regenerator2.default.wrap(function _callee$(_context) {
                    while (1) {
                        switch (_context.prev = _context.next) {
                            case 0:
                                if (!(this.permits > 0)) {
                                    _context.next = 3;
                                    break;
                                }

                                this.permits -= 1;
                                return _context.abrupt("return", _promise2.default.resolve(true));

                            case 3:
                                return _context.abrupt("return", new _promise2.default(function (resolver) {
                                    return _this.promiseResolverQueue.push(resolver);
                                }));

                            case 4:
                            case "end":
                                return _context.stop();
                        }
                    }
                }, _callee, this);
            }));
        }
        /**
         * Same as wait except the promise returned gets resolved with false if no permit becomes available in time.
         * @param {number} milliseconds The time spent waiting before the wait is aborted.
         * @return {Promise} A promise that gets resolved with true when execution is allowed to proceed or
         * false if the time given elapses before a permit becomes available.
         */

    }, {
        key: "waitFor",
        value: function waitFor(milliseconds) {
            return __awaiter(this, void 0, void 0, _regenerator2.default.mark(function _callee2() {
                var _this2 = this;

                var resolver, promise;
                return _regenerator2.default.wrap(function _callee2$(_context2) {
                    while (1) {
                        switch (_context2.prev = _context2.next) {
                            case 0:
                                if (!(this.permits > 0)) {
                                    _context2.next = 3;
                                    break;
                                }

                                this.permits -= 1;
                                return _context2.abrupt("return", _promise2.default.resolve(true));

                            case 3:
                                // We save the resolver function in the current scope so that we can resolve the promise
                                // if the time expires.
                                resolver = void 0;
                                promise = new _promise2.default(function (r) {
                                    resolver = r;
                                });
                                // The saved resolver gets added to our list of promise resolvers so that it gets a chance
                                // to be resolved as a result of a call to signal().

                                this.promiseResolverQueue.push(resolver);
                                setTimeout(function () {
                                    // We have to remove the promise resolver from our list. Resolving it twice would not be
                                    // an issue but signal() always takes the next resolver from the queue and resolves it which
                                    // would swallow a permit if we didn't remove it.
                                    var index = _this2.promiseResolverQueue.indexOf(resolver);
                                    if (index !== -1) {
                                        _this2.promiseResolverQueue.splice(index, 1);
                                    } else {}
                                    // false because the wait was unsuccessful.
                                    resolver(false);
                                }, milliseconds);
                                return _context2.abrupt("return", promise);

                            case 8:
                            case "end":
                                return _context2.stop();
                        }
                    }
                }, _callee2, this);
            }));
        }
        /**
         * Synchronous function that tries to acquire a permit and returns true if successful, false otherwise.
         * @return {boolean} Whether a permit could be acquired.
         */

    }, {
        key: "tryAcquire",
        value: function tryAcquire() {
            if (this.permits > 0) {
                this.permits -= 1;
                return true;
            }
            return false;
        }
        /**
         * Acquires and returns all permits that are currently available
         * @return {number} Acquired permits.
         */

    }, {
        key: "drainPermits",
        value: function drainPermits() {
            if (this.permits > 0) {
                var permitCount = this.permits;
                this.permits = 0;
                return permitCount;
            }
            return 0;
        }
        /**
         * Increases the number of permits by one. If there are other functions waiting, one of them will
         * continue to execute in a future iteration of the event loop.
         */

    }, {
        key: "signal",
        value: function signal() {
            this.permits += 1;
            if (this.permits > 1 && this.promiseResolverQueue.length > 0) {
                throw new Error('this._permits should never be > 0 when there is someone waiting.');
            } else if (this.permits === 1 && this.promiseResolverQueue.length > 0) {
                // If there is someone else waiting, immediately consume the permit that was released
                // at the beginning of this function and let the waiting function resume.
                this.permits -= 1;
                // promiseResolverQueue is an array of functions
                this.promiseResolverQueue.shift()(true);
            }
        }
        /**
         * Schedules func to be called once a permit becomes available.
         * @param {function} func - The function to be executed.
         * @return {any} A promise that gets resolved with the return value of the function..
         */

    }, {
        key: "execute",
        value: function execute(func) {
            return __awaiter(this, void 0, void 0, _regenerator2.default.mark(function _callee3() {
                return _regenerator2.default.wrap(function _callee3$(_context3) {
                    while (1) {
                        switch (_context3.prev = _context3.next) {
                            case 0:
                                _context3.next = 2;
                                return this.wait();

                            case 2:
                                _context3.prev = 2;
                                _context3.next = 5;
                                return func();

                            case 5:
                                return _context3.abrupt("return", _context3.sent);

                            case 6:
                                _context3.prev = 6;

                                this.signal();
                                return _context3.finish(6);

                            case 9:
                            case "end":
                                return _context3.stop();
                        }
                    }
                }, _callee3, this, [[2,, 6, 9]]);
            }));
        }
    }]);
    return Semaphore;
}();

exports.default = Semaphore;